import pytest

from piptools.exceptions import NoCandidateFound
from piptools.resolver import combine_install_requirements


@pytest.mark.parametrize(
    ("input", "expected", "prereleases", "unsafe_constraints"),
    (
        (tup + (False, set())[len(tup) - 2 :])  # noqa: E203
        for tup in [
            (["Django"], ["django==1.8"]),
            (
                ["Flask"],
                [
                    "flask==0.10.1",
                    "itsdangerous==0.24 (from flask==0.10.1)",
                    "markupsafe==0.23 (from jinja2==2.7.3->flask==0.10.1)",
                    "jinja2==2.7.3 (from flask==0.10.1)",
                    "werkzeug==0.10.4 (from flask==0.10.1)",
                ],
            ),
            (["Jinja2", "markupsafe"], ["jinja2==2.7.3", "markupsafe==0.23"]),
            # We should return a normal release version if prereleases is False
            (["SQLAlchemy"], ["sqlalchemy==0.9.9"]),
            # We should return the prerelease version if prereleases is True
            (["SQLAlchemy"], ["sqlalchemy==1.0.0b5"], True),
            # Ipython has extras available, but we don't require them in this test
            (
                ["ipython"],
                ["ipython==2.1.0", "gnureadline==6.3.3 (from ipython==2.1.0)"],
            ),
            # We should get dependencies for extras
            (
                ["ipython[notebook]"],
                [
                    "ipython[notebook]==2.1.0",
                    "pyzmq==2.1.12 (from ipython[notebook]==2.1.0)",
                    "jinja2==2.7.3 (from ipython[notebook]==2.1.0)",
                    "tornado==3.2.2 (from ipython[notebook]==2.1.0)",
                    "markupsafe==0.23 (from jinja2==2.7.3->ipython[notebook]==2.1.0)",
                    "gnureadline==6.3.3 (from ipython[notebook]==2.1.0)",
                ],
            ),
            # We should get dependencies for multiple extras
            (
                ["ipython[notebook,nbconvert]"],
                [
                    # Note that the extras should be sorted
                    "ipython[nbconvert,notebook]==2.1.0",
                    "pyzmq==2.1.12 (from ipython[nbconvert,notebook]==2.1.0)",
                    "jinja2==2.7.3 (from ipython[nbconvert,notebook]==2.1.0)",
                    "tornado==3.2.2 (from ipython[nbconvert,notebook]==2.1.0)",
                    (
                        "markupsafe==0.23 "
                        "(from jinja2==2.7.3->ipython[nbconvert,notebook]==2.1.0)"
                    ),
                    "gnureadline==6.3.3 (from ipython[nbconvert,notebook]==2.1.0)",
                    "pygments==1.5 (from ipython[nbconvert,notebook]==2.1.0)",
                    "sphinx==0.3 (from ipython[nbconvert,notebook]==2.1.0)",
                ],
            ),
            # We must take the union of all extras
            (
                ["ipython[notebook]", "ipython[nbconvert]"],
                [
                    # Note that the extras should be sorted
                    "ipython[nbconvert,notebook]==2.1.0",
                    "pyzmq==2.1.12 (from ipython[nbconvert,notebook]==2.1.0)",
                    "jinja2==2.7.3 (from ipython[nbconvert,notebook]==2.1.0)",
                    "tornado==3.2.2 (from ipython[nbconvert,notebook]==2.1.0)",
                    (
                        "markupsafe==0.23 "
                        "(from jinja2==2.7.3->ipython[nbconvert,notebook]==2.1.0)"
                    ),
                    "gnureadline==6.3.3 (from ipython[nbconvert,notebook]==2.1.0)",
                    "pygments==1.5 (from ipython[nbconvert,notebook]==2.1.0)",
                    "sphinx==0.3 (from ipython[nbconvert,notebook]==2.1.0)",
                ],
            ),
            # We must remove child dependencies from result if parent
            # is removed (e.g. vine from amqp>=2.0)
            # See: GH-370
            # because of upated dependencies in the test index, we need to pin celery
            # in order to reproduce vine removal (because it was readded in later
            # releases)
            (
                ["celery<=3.1.23", "librabbitmq"],
                [
                    "amqp==1.4.9 (from librabbitmq==1.6.1)",
                    "anyjson==0.3.3 (from kombu==3.0.35->celery==3.1.23)",
                    "billiard==3.5.0.2 (from celery==3.1.23)",
                    "celery==3.1.23",
                    "kombu==3.0.35 (from celery==3.1.23)",
                    "librabbitmq==1.6.1",
                    "pytz==2016.4 (from celery==3.1.23)",
                ],
            ),
            # Support specifying loose top-level requirements that could also appear as
            # pinned subdependencies.
            (
                ["billiard", "celery", "fake-piptools-test-with-pinned-deps"],
                [
                    "amqp==1.4.9 (from kombu==3.0.35->celery==3.1.18)",
                    "anyjson==0.3.3 (from kombu==3.0.35->celery==3.1.18)",
                    "billiard==3.3.0.23",
                    "celery==3.1.18",  # this is pinned from test subdependency
                    "fake-piptools-test-with-pinned-deps==0.1",
                    "kombu==3.0.35 (from celery==3.1.18)",
                    "pytz==2016.4 (from celery==3.1.18)",
                ],
            ),
            # Exclude package dependcy of setuptools as it is unsafe.
            (
                ["html5lib"],
                ["html5lib==0.999999999"],
                False,
                {"setuptools==35.0.0 (from html5lib==0.999999999)"},
            ),
            # We shouldn't include irrelevant pip constraints
            # See: GH-471
            (
                ["Flask", ("click", True), ("itsdangerous", True)],
                [
                    "flask==0.10.1",
                    "itsdangerous==0.24",
                    "markupsafe==0.23 (from jinja2==2.7.3->flask==0.10.1)",
                    "jinja2==2.7.3 (from flask==0.10.1)",
                    "werkzeug==0.10.4 (from flask==0.10.1)",
                ],
            ),
            # Unsafe dependencies should be filtered
            (
                ["setuptools==35.0.0", "anyjson==0.3.3"],
                ["anyjson==0.3.3"],
                False,
                {"setuptools==35.0.0"},
            ),
            (
                ["fake-piptools-test-with-unsafe-deps==0.1"],
                ["fake-piptools-test-with-unsafe-deps==0.1"],
                False,
                {
                    (
                        "setuptools==34.0.0 (from "
                        "fake-piptools-test-with-unsafe-deps==0.1)"
                    ),
                    (
                        "appdirs==1.4.9 (from "
                        "setuptools==34.0.0->fake-piptools-test-with-unsafe-deps==0.1)"
                    ),
                    (
                        "packaging==16.8 (from "
                        "setuptools==34.0.0->fake-piptools-test-with-unsafe-deps==0.1)"
                    ),
                },
            ),
        ]
    ),
)
def test_resolver(
    resolver, from_line, input, expected, prereleases, unsafe_constraints
):
    input = [line if isinstance(line, tuple) else (line, False) for line in input]
    input = [from_line(req[0], constraint=req[1]) for req in input]
    resolver = resolver(input, prereleases=prereleases)
    output = resolver.resolve()
    output = {str(line) for line in output}
    assert output == {str(line) for line in expected}
    assert {str(line) for line in resolver.unsafe_constraints} == unsafe_constraints


@pytest.mark.parametrize(
    ("input", "expected", "prereleases"),
    (
        (tup + (False,))[:3]
        for tup in [
            (
                ["setuptools==34.0.0"],
                [
                    "appdirs==1.4.9 (from setuptools==34.0.0)",
                    "packaging==16.8 (from setuptools==34.0.0)",
                    "setuptools==34.0.0",
                ],
            ),
            (
                ["fake-piptools-test-with-unsafe-deps==0.1"],
                [
                    (
                        "appdirs==1.4.9 (from "
                        "setuptools==34.0.0->fake-piptools-test-with-unsafe-deps==0.1)"
                    ),
                    (
                        "setuptools==34.0.0 "
                        "(from fake-piptools-test-with-unsafe-deps==0.1)"
                    ),
                    (
                        "packaging==16.8 (from "
                        "setuptools==34.0.0->fake-piptools-test-with-unsafe-deps==0.1)"
                    ),
                    "fake-piptools-test-with-unsafe-deps==0.1",
                ],
            ),
        ]
    ),
)
def test_resolver__allows_unsafe_deps(
    resolver, from_line, input, expected, prereleases
):
    input = [line if isinstance(line, tuple) else (line, False) for line in input]
    input = [from_line(req[0], constraint=req[1]) for req in input]
    output = resolver(input, prereleases=prereleases, allow_unsafe=True).resolve()
    output = {str(line) for line in output}
    assert output == {str(line) for line in expected}


def test_resolver__max_number_rounds_reached(resolver, from_line):
    """
    Resolver should raise an exception if max round has been reached.
    """
    input = [from_line("django")]
    with pytest.raises(RuntimeError, match="after 0 rounds of resolving"):
        resolver(input).resolve(max_rounds=0)


def test_iter_dependencies(resolver, from_line):
    """
    Dependencies should be pinned or editable.
    """
    ireq = from_line("django>=1.8")
    res = resolver([])

    with pytest.raises(
        TypeError, match="Expected pinned or editable requirement, got django>=1.8"
    ):
        next(res._iter_dependencies(ireq))


def test_combine_install_requirements(from_line):
    celery30 = from_line("celery>3.0", comes_from="-r requirements.in")
    celery31 = from_line("celery==3.1.1", comes_from=from_line("fake-package"))
    celery32 = from_line("celery<3.2")

    combined = combine_install_requirements([celery30, celery31])
    assert combined.comes_from == celery31.comes_from  # shortest string
    assert set(combined._source_ireqs) == {celery30, celery31}
    assert str(combined.req.specifier) == "==3.1.1,>3.0"

    combined_all = combine_install_requirements([celery32, combined])
    assert combined_all.comes_from is None
    assert set(combined_all._source_ireqs) == {celery30, celery31, celery32}
    assert str(combined_all.req.specifier) == "<3.2,==3.1.1,>3.0"


def test_compile_failure_shows_provenance(resolver, from_line):
    """
    Provenance of conflicting dependencies should be printed on failure.
    """
    requirements = [
        from_line("fake-piptools-test-with-pinned-deps==0.1"),
        from_line("celery>3.2"),
    ]

    with pytest.raises(NoCandidateFound) as err:
        resolver(requirements).resolve()
    lines = str(err.value).splitlines()
    assert (
        lines[-2].strip()
        == "celery==3.1.18 (from fake-piptools-test-with-pinned-deps==0.1)"
    )
    assert lines[-1].strip() == "celery>3.2"



from pytest import fixture, mark, raises

from piptools._compat import FormatControl
from piptools.scripts.compile import cli
from piptools.utils import comment
from piptools.writer import (
    MESSAGE_UNHASHED_PACKAGE,
    MESSAGE_UNSAFE_PACKAGES,
    MESSAGE_UNSAFE_PACKAGES_UNPINNED,
    OutputWriter,
)


@fixture
def writer(tmpdir_cwd):
    with open("src_file", "w"), open("src_file2", "w"):
        pass

    cli_args = [
        "--dry-run",
        "--output-file",
        "requirements.txt",
        "src_file",
        "src_file2",
    ]

    with cli.make_context("pip-compile", cli_args) as ctx:
        writer = OutputWriter(
            src_files=["src_file", "src_file2"],
            dst_file=ctx.params["output_file"],
            click_ctx=ctx,
            dry_run=True,
            emit_header=True,
            emit_index=True,
            emit_trusted_host=True,
            annotate=True,
            generate_hashes=False,
            default_index_url=None,
            index_urls=[],
            trusted_hosts=[],
            format_control=FormatControl(set(), set()),
            allow_unsafe=False,
            find_links=[],
        )
        yield writer


def test_format_requirement_annotation_editable(from_editable, writer):
    # Annotations are printed as comments at a fixed column
    ireq = from_editable("git+git://fake.org/x/y.git#egg=y")
    reverse_dependencies = {"y": ["xyz"]}

    assert writer._format_requirement(
        ireq, reverse_dependencies, primary_packages=[]
    ) == "-e git+git://fake.org/x/y.git#egg=y  " + comment("# via xyz")


def test_format_requirement_annotation(from_line, writer):
    ireq = from_line("test==1.2")
    reverse_dependencies = {"test": ["xyz"]}

    assert writer._format_requirement(
        ireq, reverse_dependencies, primary_packages=[]
    ) == "test==1.2                 " + comment("# via xyz")


def test_format_requirement_annotation_lower_case(from_line, writer):
    ireq = from_line("Test==1.2")
    reverse_dependencies = {"test": ["xyz"]}

    assert writer._format_requirement(
        ireq, reverse_dependencies, primary_packages=[]
    ) == "test==1.2                 " + comment("# via xyz")


def test_format_requirement_not_for_primary(from_line, writer):
    "Primary packages should not get annotated."
    ireq = from_line("test==1.2")
    reverse_dependencies = {"test": ["xyz"]}

    assert (
        writer._format_requirement(
            ireq, reverse_dependencies, primary_packages=["test"]
        )
        == "test==1.2"
    )


def test_format_requirement_not_for_primary_lower_case(from_line, writer):
    "Primary packages should not get annotated."
    ireq = from_line("Test==1.2")
    reverse_dependencies = {"test": ["xyz"]}

    assert (
        writer._format_requirement(
            ireq, reverse_dependencies, primary_packages=["test"]
        )
        == "test==1.2"
    )


def test_format_requirement_environment_marker(from_line, writer):
    "Environment markers should get passed through to output."
    ireq = from_line(
        'test ; python_version == "2.7" and platform_python_implementation == "CPython"'
    )
    reverse_dependencies = set()

    result = writer._format_requirement(
        ireq, reverse_dependencies, primary_packages=["test"], marker=ireq.markers
    )
    assert (
        result == 'test ; python_version == "2.7" and '
        'platform_python_implementation == "CPython"'
    )


@mark.parametrize(("allow_unsafe",), [(True,), (False,)])
def test_iter_lines__unsafe_dependencies(writer, from_line, allow_unsafe):
    writer.allow_unsafe = allow_unsafe
    output = "\n".join(
        writer._iter_lines([from_line("test==1.2")], [from_line("setuptools")])
    )
    assert (
        "\n".join(
            [
                "test==1.2",
                "",
                MESSAGE_UNSAFE_PACKAGES,
                "setuptools" if allow_unsafe else comment("# setuptools"),
            ]
        )
        in output
    )


def test_iter_lines__unsafe_with_hashes(writer, from_line):
    writer.allow_unsafe = False
    ireqs = [from_line("test==1.2")]
    unsafe_ireqs = [from_line("setuptools")]
    hashes = {ireqs[0]: {"FAKEHASH"}, unsafe_ireqs[0]: set()}
    output = "\n".join(writer._iter_lines(ireqs, unsafe_ireqs, hashes=hashes))
    assert (
        "\n".join(
            [
                "test==1.2 \\",
                "    --hash=FAKEHASH",
                "",
                MESSAGE_UNSAFE_PACKAGES_UNPINNED,
                comment("# setuptools"),
            ]
        )
        in output
    )


def test_iter_lines__hash_missing(writer, from_line):
    ireqs = [from_line("test==1.2"), from_line("file:///example/#egg=example")]
    hashes = {ireqs[0]: {"FAKEHASH"}, ireqs[1]: set()}
    output = "\n".join(writer._iter_lines(ireqs, hashes=hashes))
    assert (
        "\n".join(
            [MESSAGE_UNHASHED_PACKAGE, "file:///example/#egg=example", "test==1.2"]
        )
        in output
    )


def test_write_header(writer):
    expected = map(
        comment,
        [
            "#",
            "# This file is autogenerated by pip-compile",
            "# To update, run:",
            "#",
            "#    pip-compile --output-file={} src_file src_file2".format(
                writer.click_ctx.params["output_file"].name
            ),
            "#",
        ],
    )
    assert list(writer.write_header()) == list(expected)


def test_write_header_custom_compile_command(writer, monkeypatch):
    monkeypatch.setenv("CUSTOM_COMPILE_COMMAND", "./pipcompilewrapper")
    expected = map(
        comment,
        [
            "#",
            "# This file is autogenerated by pip-compile",
            "# To update, run:",
            "#",
            "#    ./pipcompilewrapper",
            "#",
        ],
    )
    assert list(writer.write_header()) == list(expected)


def test_write_header_no_emit_header(writer):
    """
    There should not be headers if emit_header is False
    """
    writer.emit_header = False

    with raises(StopIteration):
        next(writer.write_header())


def test_write_format_controls(writer):
    """
    Tests --no-binary/--only-binary options.
    """

    writer.format_control = FormatControl(
        no_binary=["psycopg2", "click"], only_binary=["pytz", "django"]
    )
    lines = list(writer.write_format_controls())

    assert "--no-binary psycopg2" in lines
    assert "--no-binary click" in lines

    assert "--only-binary pytz" in lines
    assert "--only-binary django" in lines


@mark.parametrize(
    ("index_urls", "expected_lines"),
    (
        # No urls - no options
        [[], []],
        # Single URL should be index-url
        [["https://index-url.com"], ["--index-url https://index-url.com"]],
        # First URL should be index-url, the others should be extra-index-url
        [
            [
                "https://index-url1.com",
                "https://index-url2.com",
                "https://index-url3.com",
            ],
            [
                "--index-url https://index-url1.com",
                "--extra-index-url https://index-url2.com",
                "--extra-index-url https://index-url3.com",
            ],
        ],
        # If a first URL equals to the default URL, the the index url must not be set
        # and the others should be extra-index-url
        [
            [
                "https://default-index-url.com",
                "https://index-url1.com",
                "https://index-url2.com",
            ],
            [
                "--extra-index-url https://index-url1.com",
                "--extra-index-url https://index-url2.com",
            ],
        ],
        # Ignore URLs equal to the default index-url
        # (note: the previous case is exception)
        [
            [
                "https://index-url1.com",
                "https://default-index-url.com",
                "https://index-url2.com",
            ],
            [
                "--index-url https://index-url1.com",
                "--extra-index-url https://index-url2.com",
            ],
        ],
        # Ignore URLs equal to the default index-url
        [["https://default-index-url.com", "https://default-index-url.com"], []],
        # All URLs must be deduplicated
        [
            [
                "https://index-url1.com",
                "https://index-url1.com",
                "https://index-url2.com",
            ],
            [
                "--index-url https://index-url1.com",
                "--extra-index-url https://index-url2.com",
            ],
        ],
    ),
)
def test_write_index_options(writer, index_urls, expected_lines):
    """
    Test write_index_options method.
    """
    writer.index_urls = index_urls
    writer.default_index_url = "https://default-index-url.com"
    assert list(writer.write_index_options()) == expected_lines


def test_write_index_options_no_emit_index(writer):
    """
    There should not be --index-url/--extra-index-url options if emit_index is False.
    """
    writer.emit_index = False
    with raises(StopIteration):
        next(writer.write_index_options())


@mark.parametrize(
    "find_links, expected_lines",
    (
        [[], []],
        [["./foo"], ["--find-links ./foo"]],
        [["./foo", "./bar"], ["--find-links ./foo", "--find-links ./bar"]],
    ),
)
def test_write_find_links(writer, find_links, expected_lines):
    """
    Test write_find_links method.
    """
    writer.find_links = find_links
    assert list(writer.write_find_links()) == expected_lines
